import importlib.resources

from ....base_classes import InterpolatedDataModel
from .reference import CIEModel203


class DirectTransmission(InterpolatedDataModel, CIEModel203):
    def __init__(self):
        datafile = importlib.resources.path(
            __package__, "direct_transmission-human_eye-cie203_2012.txt"
        )
        super().__init__("direct ocular transmission", "", "wavelength", "um", datafile)



from ....units import Q_

lambda_0_squared = Q_(0.028, "um^2")


def alpha_1(l: Q_) -> Q_:
    """
    alpha_1 used by Navarro, defined immediately after equation 2.

    alpha_1 - alpha_4 all have the same pattern

    alpha_x(lambda) = A + B*lambda**2 + C / (lmabda**2 - lambda_0**2) + D / (lmabda**2 - lambda_0**2)**2
    """
    l = Q_(l)

    # NOTE: there is an error in the Navarro paper, they use the coefficientes from Herzberger,
    # but there is a a typo in B.
    A = Q_(0.66147196, "")
    B = Q_(-0.40352796, "1/um^2")  # <<<< Navarro paper has -0.040352796
    C = Q_(-0.2804679, "um^2")
    D = Q_(0.03385979, "um^4")

    return (
        A
        + B * l**2
        + C / (l**2 - lambda_0_squared)
        + D / (l**2 - lambda_0_squared) ** 2
    ).to("")


def alpha_2(l: Q_) -> Q_:
    l = Q_(l)

    A = Q_(-4.20146383, "")
    B = Q_(2.73508956, "1/um^2")
    C = Q_(1.50543784, "um^2")
    D = Q_(-0.11593235, "um^4")

    return (
        A
        + B * l**2
        + C / (l**2 - lambda_0_squared)
        + D / (l**2 - lambda_0_squared) ** 2
    ).to("")


def alpha_3(l: Q_) -> Q_:
    l = Q_(l)

    A = Q_(6.29834237, "")
    B = Q_(-4.69409935, "1/um^2")
    C = Q_(-1.5750865, "um^2")
    D = Q_(0.10293038, "um^4")

    return (
        A
        + B * l**2
        + C / (l**2 - lambda_0_squared)
        + D / (l**2 - lambda_0_squared) ** 2
    ).to("")


def alpha_4(l: Q_) -> Q_:
    l = Q_(l)

    # NOTE: there is an another error in the Navarro paper, they use the coefficientes from Hrzberger,
    # but there is a a typo in A.
    A = Q_(-1.75835059, "")  # <<<< Navarro paper has +1.75835059
    B = Q_(2.36253794, "1/um^2")
    C = Q_(0.35011657, "um^2")
    D = Q_(-0.02085782, "um^4")

    return (
        A
        + B * l**2
        + C / (l**2 - lambda_0_squared)
        + D / (l**2 - lambda_0_squared) ** 2
    ).to("")


def n(n_double_star: Q_, n_F: Q_, n_c: Q_, n_star: Q_, l: Q_) -> Q_:
    return (
        alpha_1(l) * n_double_star
        + alpha_2(l) * n_F
        + alpha_3(l) * n_c
        + alpha_4(l) * n_star
    )



import numpy

from ....units import Q_
from .reference import SchulmeisterModel


class Melanin(SchulmeisterModel):
    def __init__(self):
        self.argument_units = "nm"
        self.value_units = "1/cm"
        pass

    def __call__(self, wavelength: str | Q_):
        wavelength = Q_(wavelength)
        wavelength.ito("nm")
        wavelength = wavelength.magnitude
        val = 3.85e14 * wavelength ** (-4.2)

        return Q_(val, "1/cm")



import numpy

from ....units import Q_
from .blood import Blood
from .melanin import Melanin
from .reference import SchulmeisterModel


class Choroid(SchulmeisterModel):
    def __init__(self):
        super().__init__()
        self.blood = Blood()
        self.melanin = Melanin()

    def __call__(self, wavelength: str | Q_):
        wavelength = Q_(wavelength)
        wavelength.ito("nm")
        val = 0.114 * self.melanin(wavelength) + 0.3 * self.blood(wavelength)

        return Q_(val, "1/cm")



import numpy

from ....units import Q_
from .reference import SchulmeisterModel


class Blood(SchulmeisterModel):
    def __init__(self):
        self.argument_units = "nm"
        self.value_units = "1/cm"
        pass

    def __call__(self, wavelength: str | Q_):
        wavelength = Q_(wavelength)
        wavelength.ito("nm")
        wavelength = wavelength.magnitude
        val = 1.404e7 * numpy.exp(-0.024 * wavelength)
        val += 2160 * numpy.exp(-0.0055 * (417 - wavelength) ** 2)
        val += 244 * numpy.exp(-0.0022 * (543 - wavelength) ** 2)
        val += 253 * numpy.exp(-0.0065 * (577 - wavelength) ** 2)
        val += 7 * numpy.exp(-0.00004 * (920 - wavelength) ** 2)

        return Q_(val, "1/cm")



class UndefinedProperty(RuntimeError):
    def __init__(self, property_name: str, argument_name: str):
        super().__init__(
            f"Property ({property_name}) is not defined for argument ({argument_name}). I.e. the argument is outside the valid range for the property in this dataset."
        )



from ....base_classes import ReferencedDataModel


class CIEModel203(ReferencedDataModel):
    reference = """
@TechReport{lund-2012,
author      = {D. Lund and J. Marshall and J. Mellerio and T. Okuno and K. Schulmeister and D. Sliney and P. S\"{o}derberg and B. Stuck and D. {van Norren} and J. Zuclich},
title       = {A computerized approach to transmission and absorption characteristics of the human eye},
institution = {International Commission on Illumination},
year        = {2012},
url         = {http://cie.co.at/publications/computerized-approach-transmission-and-absorption-characteristics-human-eye},
}
"""

    def __init__(self):
        pass



import numpy

from ....units import Q_
from .reference import NavarroModel
from .utils import n


class Lens(NavarroModel):
    def __init__(self):
        pass

    def __call__(self, wavelength: str | Q_):
        n_double_star = Q_(1.4492,'')
        n_F = Q_(1.42625,'')
        n_c = Q_(1.4175,'')
        n_star = Q_(1.4097,'')

        wavelength = Q_(wavelength)

        val = n(n_double_star,n_F,n_c,n_star,wavelength)

        return val



from .reference import SchulmeisterModel
from ....units import Q_
import numpy


class MacularPigment(SchulmeisterModel):
    def __init__(self):
        self.argument_units = "nm"
        self.value_units = "1/cm"
        pass
    

    def __call__(self,wavelength: str|Q_):
        wavelength = Q_(wavelength)
        wavelength.ito("nm")
        wavelength = wavelength.magnitude
        val = 0
        val += 0.32*numpy.exp(-0.0012*(436-wavelength)**2)
        val += 0.32*numpy.exp(-0.0012*(480-wavelength)**2)
        val -= 0.123*numpy.exp(-0.0012*(458-wavelength)**2)
        val += 0.12042*numpy.exp(-0.006*(457-wavelength)**2)
        val *= 143/0.35

        return Q_(val,'1/cm')




import importlib.resources

from ....base_classes import InterpolatedDataModel
from .reference import MainsterModel


class Transmission(InterpolatedDataModel, MainsterModel):
    def __init__(self):
        datafile = importlib.resources.path(
            __package__, "transmission-human_eye-mainster.txt"
        )
        super().__init__("ocular transmission", "", "wavelength", "nm", datafile)



import importlib.resources

from ....base_classes import InterpolatedDataModel
from .reference import MainsterModel


class Choroid(InterpolatedDataModel, MainsterModel):
    def __init__(self):
        datafile = importlib.resources.path(__package__, "mua-mainster-choroid.txt")
        super().__init__("absorption coefficient", "1/cm", "wavelength", "nm", datafile)



import numpy
import scipy

from .exceptions import UndefinedProperty
from .units import Q_


class ReferencedDataModel:
    reference = None

    def __init__(self):
        pass

    def get_reference(self):
        return self.reference


class InterpolatedDataModel:
    def __init__(self, vname: str, vunit: str, aname: str, aunit: str, filename=None):
        self.filename = filename
        self.interpolator = None
        self.value_name = vname
        self.value_unit = vunit
        self.argument_name = aname
        self.argument_unit = aunit
        if self.filename is not None:
            self.load_data_from_file(self.filename)

    def load_data_from_file(self, filename):
        data = numpy.loadtxt(filename)
        self.interpolator = scipy.interpolate.interp1d(data[:, 0], data[:, 1])

    def __call__(self, arg: str | Q_):
        arg = Q_(arg)
        arg.ito(self.argument_unit)
        try:
            val = Q_(self.interpolator(arg.magnitude), self.value_unit)
        except ValueError as e:
            raise UndefinedProperty(self.value_name, f"{self.argument_name} = {arg}")
        return val



from ....base_classes import ReferencedDataModel


class SchulmeisterModel(ReferencedDataModel):
    reference = """
@Article{schulmeister-2017,
  author  = {Mathieu Jean and Karl Schulmeister},
  title   = {Validation of a computer model to predict laser induced retinal injury thresholds},
  journal = {Journal of Laser Applications},
  year    = {2017},
  url     = {https://lia.scitation.org/doi/10.2351/1.4997831},
}
"""

    def __init__(self):
        pass



from ....base_classes import ReferencedDataModel


class MainsterModel(ReferencedDataModel):
    reference = """
@Article{mainster-1970,
  Title                    = {Retinal-Temperature Increases Produced by Intense Light Sources},
  Author                   = {M. A. Mainster and T. J. White and J. H. Tips and P. W. Wilson},
  Journal                  = {Journal of the Optical Society of America},
  Year                     = {1970},
  Number                   = {2},
  Pages                    = {264--270},
  Volume                   = {60},
}
"""

    def __init__(self):
        pass



import numpy

from ....units import Q_
from ..cie203.total_transmission import TotalTransmission
from .reference import SchulmeisterModel


class EffectiveTransmission(SchulmeisterModel):
    def __init__(self):
        self.total_transmission = TotalTransmission()

    def g(self, wavelength: str | Q_):
        wavelength = Q_(wavelength)
        wavelength.ito("nm")
        A = 0.5
        B = Q_(883, "nm")
        return A * numpy.exp(-wavelength / B)

    def h(self, spot_size: str | Q_):
        spot_size = Q_(spot_size)
        spot_size.ito("um")
        C = Q_(600, "um")
        return numpy.exp(-spot_size / C)

    def __call__(self, wavelength: str | Q_, spot_size: str | Q_):
        return self.total_transmission(wavelength) * (
            1 - self.g(wavelength) * self.h(spot_size)
        )



from ....base_classes import ReferencedDataModel


class NavarroModel(ReferencedDataModel):
    reference = """
@article{navarro-1985,
  title={Accommodation-dependent model of the human eye with aspherics.},
  author={Rafael Fonolla Navarro and Javier Santamar{\'i}a and Juli{\'a}n Besc{\'o}s},
  journal={Journal of the Optical Society of America. A, Optics and image science},
  year={1985},
  volume={2 8},
  pages={ 1273-81 },
  url={https://api.semanticscholar.org/CorpusID:14921290}
}
"""

    def __init__(self):
        pass
